/*  Copyright (C) 2007-2009  Pako, Inc.
 *  Authors: Jeff Arnold, Anders Kaseorg, Tim Abbott
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
 *  02110-1301, USA.
 */

#ifdef PAKO_STANDALONE
#include "pako.h"
#else
#include <linux/pako.h>
#endif

extern struct pako_reloc pako_relocs[], pako_relocs_end[];
extern struct pako_section pako_sections[], pako_sections_end[];
extern struct pako_symbol pako_symbols[], pako_symbols_end[];
extern struct pako_patch pako_patches[], pako_patches_end[];
extern const typeof(int (*)(void)) pako_call_pre_apply[],
    pako_call_pre_apply_end[], pako_call_check_apply[],
    pako_call_check_apply_end[];
extern const typeof(void (*)(void)) pako_call_apply[],
    pako_call_apply_end[], pako_call_post_apply[],
    pako_call_post_apply_end[], pako_call_fail_apply[],
    pako_call_fail_apply_end[];
extern const typeof(int (*)(void)) pako_call_pre_reverse[],
    pako_call_pre_reverse_end[], pako_call_check_reverse[],
    pako_call_check_reverse_end[];
extern const typeof(void (*)(void)) pako_call_reverse[],
    pako_call_reverse_end[], pako_call_post_reverse[],
    pako_call_post_reverse_end[], pako_call_fail_reverse[],
    pako_call_fail_reverse_end[];

#ifdef PAKO_STANDALONE
extern struct pako_system_map pako_system_map[], pako_system_map_end[];
#endif /* PAKO_STANDALONE */

#define change PAKO_UNIQ(change)
struct pako_mod_change change = {
	.name = "pako_" __stringify(PAKO_MID),
	.kid = __stringify(PAKO_KID),
	.target_name = __stringify(PAKO_TARGET),
#ifdef PAKO_STANDALONE
	.map_printk = MAP_PRINTK,
#endif /* PAKO_STANDALONE */
	.new_code_mod = THIS_MODULE,
	.new_code = {
		.relocs = pako_relocs,
		.relocs_end = pako_relocs_end,
		.sections = pako_sections,
		.sections_end = pako_sections_end,
		.symbols = pako_symbols,
		.symbols_end = pako_symbols_end,
#ifdef PAKO_STANDALONE
		.system_map = pako_system_map,
		.system_map_end = pako_system_map_end,
#endif /* PAKO_STANDALONE */
	},
	.patches = pako_patches,
	.patches_end = pako_patches_end,
	.hooks = {
		[KS_APPLY] = {
			.pre = pako_call_pre_apply,
			.pre_end = pako_call_pre_apply_end,
			.check = pako_call_check_apply,
			.check_end = pako_call_check_apply_end,
			.intra = pako_call_apply,
			.intra_end = pako_call_apply_end,
			.post = pako_call_post_apply,
			.post_end = pako_call_post_apply_end,
			.fail = pako_call_fail_apply,
			.fail_end = pako_call_fail_apply_end,
		},
		[KS_REVERSE] = {
			.pre = pako_call_pre_reverse,
			.pre_end = pako_call_pre_reverse_end,
			.check = pako_call_check_reverse,
			.check_end = pako_call_check_reverse_end,
			.intra = pako_call_reverse,
			.intra_end = pako_call_reverse_end,
			.post = pako_call_post_reverse,
			.post_end = pako_call_post_reverse_end,
			.fail = pako_call_fail_reverse,
			.fail_end = pako_call_fail_reverse_end,
		},
	},
};
EXPORT_SYMBOL_GPL(change);

static int init_new_code(void)
{
	return 0;
}

static void cleanup_new_code(void)
{
	cleanup_pako_mod_change(&change);
}

module_init(init_new_code);
module_exit(cleanup_new_code);

MODULE_AUTHOR("Pako, Inc.");
MODULE_DESCRIPTION("Pako rebootless update new code module");
#ifdef PAKO_VERSION
MODULE_VERSION(PAKO_VERSION);
#endif
MODULE_LICENSE("GPL v2");
