#ifndef _PAKO_PATCH_H
#define _PAKO_PATCH_H

#define PAKO_OPTION_ASSUME_RODATA	0
#define PAKO_OPTION_MATCH_DATA_EARLY	1

struct pako_option {
	int type;
	const void *target;
} __attribute__((packed));

#ifdef __KERNEL__

#include <linux/gfp.h>
#include <linux/stringify.h>

#ifndef __used
#define __used __attribute_used__
#endif

#define pako_call_int(name, fn) \
	static typeof(int (*)(void)) __pako_##name##_##fn __used \
	__attribute__((__section__(".pako_call_" #name))) = fn

#define pako_call_void(name, fn) \
	static typeof(void (*)(void)) __pako_##name##_##fn __used \
	__attribute__((__section__(".pako_call_" #name))) = fn

#define pako_pre_apply(fn) pako_call_int(pre_apply, fn)
#define pako_check_apply(fn) pako_call_int(check_apply, fn)
#define pako_apply(fn) pako_call_void(apply, fn)
#define pako_post_apply(fn) pako_call_void(post_apply, fn)
#define pako_fail_apply(fn) pako_call_void(fail_apply, fn)

#define pako_pre_reverse(fn) pako_call_int(pre_reverse, fn)
#define pako_check_reverse(fn) pako_call_int(check_reverse, fn)
#define pako_reverse(fn) pako_call_void(reverse, fn)
#define pako_post_reverse(fn) pako_call_void(post_reverse, fn)
#define pako_fail_reverse(fn) pako_call_void(fail_reverse, fn)


#define pako_assume_rodata(obj)				\
	pako_option(PAKO_OPTION_ASSUME_RODATA, obj)

#define pako_match_data_early(obj)				\
	pako_option(PAKO_OPTION_MATCH_DATA_EARLY, obj)

#if BITS_PER_LONG == 32
#define PAKO_PTR ".long"
#elif BITS_PER_LONG == 64
#define PAKO_PTR ".quad"
#endif /* BITS_PER_LONG */

#define pako_option(num, obj)				\
	__asm__(".pushsection \".pako_options\", \"a\"\n"	\
		"\t.long " __stringify(num) "\n"		\
		"\t" PAKO_PTR " " #obj "\n"			\
		".popsection")

int init_shadow_field_type(int *shadow_key, typeof(GFP_KERNEL) gfp_flags);
void *init_shadow_field(int *shadow_key, void *obj, int size,
			typeof(GFP_KERNEL) gfp_flags);
void cleanup_shadow_field(int *shadow_key, void *obj);
void *get_shadow_field(int *shadow_key, void *obj);
void cleanup_shadow_field_type(int *shadow_key);


#define __DEFINE_SHADOW_FIELD(base_type, field_type, gfp_flags,			\
			      init_field_type_fn, init_field_fn, get_field_fn,	\
			      make_field_fn, cleanup_field_fn,			\
			      cleanup_field_type_fn, shadow_key, init_field)	\
	static int shadow_key = 0;						\
	int init_field_type_fn(void)						\
	{									\
		return init_shadow_field_type(&shadow_key, gfp_flags);		\
	}									\
	field_type *init_field_fn(base_type *obj, typeof(GFP_KERNEL) flags)	\
	{									\
		field_type *data = init_shadow_field(&shadow_key, (void *)obj,	\
						     sizeof(*data), flags);	\
		if (data != NULL)						\
			init_field(data);					\
		return data;							\
	}									\
	void cleanup_field_fn(base_type *obj)					\
	{									\
		cleanup_shadow_field(&shadow_key, obj);				\
	}									\
	field_type *get_field_fn(base_type *obj)				\
	{									\
		return get_shadow_field(&shadow_key, obj);			\
	}									\
	field_type *make_field_fn(base_type *obj, typeof(GFP_KERNEL) flags)	\
	{									\
		void *data = get_shadow_field(&shadow_key, (void *)obj);	\
		if (data == NULL)						\
			data = init_field_fn(obj, flags);		\
		return data;							\
	}									\
	void cleanup_field_type_fn(void)					\
	{									\
		return cleanup_shadow_field_type(&shadow_key);			\
	}									\
	struct eat_trailing_semicolon

#define DEFINE_SHADOW_FIELD(base_type, field_type, gfp_flags, name, init_field)	\
	__DEFINE_SHADOW_FIELD(base_type, field_type, gfp_flags,			\
			      init_##name##_shadows, init_##name##_shadow,	\
			      get_##name##_shadow, make_##name##_shadow,	\
			      cleanup_##name##_shadow, cleanup_##name##_shadows,\
			      shadow_key_##name, init_field);			\
	pako_check_apply(init_##name##_shadows);				\
	pako_post_reverse(cleanup_##name##_shadows);				\
	pako_fail_apply(cleanup_##name##_shadows)

#define __DECLARE_SHADOW_FIELD(base_type, field_type, init_field_type_fn,	\
			       init_field_fn, get_field_fn, make_field_fn,	\
			       cleanup_field_fn, cleanup_field_type_fn)		\
	int init_field_type_fn(void);						\
	field_type *init_field_fn(base_type *obj, typeof(GFP_KERNEL) flags);	\
	void cleanup_field_fn(base_type *obj);					\
	field_type *get_field_fn(base_type *obj);				\
	field_type *make_field_fn(base_type *obj, typeof(GFP_KERNEL) flags);	\
	void cleanup_field_type_fn(void)

#define DECLARE_SHADOW_FIELD(base_type, field_type, name)			\
	__DECLARE_SHADOW_FIELD(base_type, field_type, init_##name##_shadows,	\
			       init_##name##_shadow, get_##name##_shadow,	\
			       make_##name##_shadow, cleanup_##name##_shadow,	\
			       cleanup_##name##_shadows)

#endif /* __KERNEL__ */

#endif /* _PAKO_PATCH_H */
